# Copyright 2024 Zsbyqx20 <112002598+Zsbyqx20@users.noreply.github.com>
# SPDX-FileCopyrightText: 2024 Zsbyqx20 <112002598+Zsbyqx20@users.noreply.github.com>
#
# SPDX-License-Identifier: Apache-2.0

from .anthropic import AnthropicClient
from .openai import OpenAIClient

__all__ = ["OpenAIClient", "AnthropicClient"]



# SPDX-FileCopyrightText: 2024 Zsbyqx20 <112002598+Zsbyqx20@users.noreply.github.com>
#
# SPDX-License-Identifier: Apache-2.0

import asyncio
import io
import time
from functools import wraps
from typing import Awaitable, Callable, ParamSpec, TypeVar

import requests
from PIL import Image

from .base import ImageFormat

RT = TypeVar("RT")
P = ParamSpec("P")


def retry(
    max_attempts: int = 3,
    delay: float = 1.0,
    backoff_factor: float = 2.0,
    skip_exceptions: tuple[type[Exception], ...] = (),
) -> Callable[[Callable[P, RT]], Callable[P, RT]]:
    def decorator(func: Callable[P, RT]) -> Callable[P, RT]:
        @wraps(func)
        def wrapper(*args: P.args, **kwargs: P.kwargs) -> RT:
            current_delay = delay
            last_exception = None

            for attempt in range(max_attempts):
                try:
                    return func(*args, **kwargs)
                except Exception as e:
                    if isinstance(e, skip_exceptions):
                        raise
                    last_exception = e
                    if attempt == max_attempts - 1:
                        raise
                    time.sleep(current_delay)
                    current_delay *= backoff_factor

            raise last_exception  # type: ignore

        return wrapper

    return decorator


def async_retry(
    max_attempts: int = 3,
    delay: float = 1.0,
    backoff_factor: float = 2.0,
    skip_exceptions: tuple[type[Exception], ...] = (),
) -> Callable[[Callable[P, Awaitable[RT]]], Callable[P, Awaitable[RT]]]:
    def decorator(func: Callable[P, Awaitable[RT]]) -> Callable[P, Awaitable[RT]]:
        @wraps(func)
        async def wrapper(*args: P.args, **kwargs: P.kwargs) -> RT:
            current_delay = delay
            last_exception = None

            for attempt in range(max_attempts):
                try:
                    return await func(*args, **kwargs)
                except Exception as e:
                    if isinstance(e, skip_exceptions):
                        raise
                    last_exception = e
                    if attempt == max_attempts - 1:
                        raise
                    await asyncio.sleep(current_delay)
                    current_delay *= backoff_factor

            raise last_exception  # type: ignore

        return wrapper

    return decorator


@retry(max_attempts=3)
def is_valid_image_url(url: str, timeout: int = 5) -> tuple[Image.Image, ImageFormat]:
    try:
        response = requests.head(url, timeout=timeout)
        content_type = response.headers.get("content-type", "")

        if not content_type.startswith("image/"):
            raise ValueError(f"Not an image. Content-Type: {content_type}")

        response = requests.get(url, timeout=timeout)
        image_data = io.BytesIO(response.content)

        img = Image.open(image_data)
        image_format = img.format.lower() if img.format else "unknown"

        img_copy = img.copy()
        img.verify()

        try:
            return img_copy, ImageFormat[image_format.upper()]
        except KeyError:
            return img_copy, ImageFormat.UNKNOWN

    except requests.ConnectionError:
        raise ConnectionError("Failed to connect to the server")
    except requests.Timeout:
        raise TimeoutError("Request timed out")
    except requests.RequestException as e:
        raise RuntimeError(f"Request error: {str(e)}")
    except (IOError, SyntaxError) as e:
        raise ValueError(f"Invalid image data: {str(e)}")



# SPDX-FileCopyrightText: 2024 Zsbyqx20 <112002598+Zsbyqx20@users.noreply.github.com>
#
# SPDX-License-Identifier: Apache-2.0

import os
from abc import ABC, abstractmethod
from enum import Enum
from typing import Any, Generic, TypeVar, Union

from dotenv import load_dotenv
from PIL import Image

load_dotenv()

T = TypeVar("T")


class ContentType(Enum):
    TEXT = "text"
    IMAGE = "image"


class Role(Enum):
    USER = "user"
    SYSTEM = "system"
    ASSISTANT = "assistant"


class ImageFormat(Enum):
    PNG = "png"
    JPEG = "jpeg"
    WEBP = "webp"
    GIF = "gif"
    UNKNOWN = "unknown"


class ImageDetail(Enum):
    HIGH = "high"
    LOW = "low"
    AUTO = "auto"


class ServiceType(Enum):
    OPENAI = "openai"
    ANTHROPIC = "anthropic"


MessageInput = Union[str, dict[str, Any], "BaseMessage", Image.Image]


class BaseMessage(Generic[T]):
    def __init__(
        self,
        content: T,
        role: Role = Role.USER,
        content_type: ContentType = ContentType.TEXT,
    ):
        self._role = role
        self._content = content
        self._content_type = content_type

    @property
    def role(self):
        return self._role

    @property
    def content(self):
        return self._content

    @property
    def content_type(self):
        return self._content_type


class BaseClient(ABC):
    def __init__(
        self,
        api_key: str | None = None,
        base_url: str | None = None,
        service_type: ServiceType = ServiceType.OPENAI,
    ):
        self.service_type = service_type
        self.api_key = api_key or os.getenv(f"{service_type.value.upper()}_API_KEY")
        self.base_url = base_url or os.getenv(f"{service_type.value.upper()}_BASE_URL")
        if self.api_key is None:
            raise ValueError(f"{service_type.value.upper()}_API_KEY is not set")
        self._init_client()

    @abstractmethod
    def _init_client(self):
        raise NotImplementedError

    @abstractmethod
    def organize_messages(self, messages: list[BaseMessage]):
        raise NotImplementedError

    def _validate_message_sequence(self, messages: list[BaseMessage]):
        system_indices = [
            i for i, msg in enumerate(messages) if msg.role == Role.SYSTEM
        ]
        if len(system_indices) > 1:
            raise ValueError("Multiple system messages are not allowed")
        if system_indices and system_indices[0] != 0:
            raise ValueError("System message must be the first message if present")
        if messages and messages[-1].role == Role.ASSISTANT:
            raise ValueError("Message sequence cannot end with an assistant message")



# SPDX-FileCopyrightText: 2024 Zsbyqx20 <112002598+Zsbyqx20@users.noreply.github.com>
#
# SPDX-License-Identifier: Apache-2.0

import base64
import io
from typing import NotRequired, TypedDict, Union

from PIL import Image

from .base import BaseMessage, ContentType, ImageDetail, ImageFormat, MessageInput, Role
from .utils import is_valid_image_url

MessageDict = TypedDict(
    "MessageDict",
    {
        "content": Union[str, Image.Image],
        "role": NotRequired[str],
        "type": NotRequired[str],
        "format": NotRequired[str],
        "detail": NotRequired[str],
    },
)


class TextMessage(BaseMessage[str]):
    pass


class ImageMessage(BaseMessage[Image.Image]):
    def __init__(
        self,
        content: Image.Image | str,
        format: ImageFormat = ImageFormat.PNG,
        detail: ImageDetail = ImageDetail.AUTO,
    ):
        if isinstance(content, str):
            content_, format_ = is_valid_image_url(content)
            content = content_
            if format_ == ImageFormat.UNKNOWN:
                format = ImageFormat.PNG
        super().__init__(content, role=Role.USER, content_type=ContentType.IMAGE)
        self._format = format
        self._detail = detail

    @property
    def content(self) -> str:
        img_byte_arr = io.BytesIO()
        self._content.save(img_byte_arr, format=self._format.value)
        img_byte_arr = img_byte_arr.getvalue()
        return base64.b64encode(img_byte_arr).decode()

    @property
    def detail(self):
        return self._detail


def convert_to_message(msg: MessageInput) -> BaseMessage:
    if isinstance(msg, str):
        return TextMessage(content=msg, role=Role.USER)

    if isinstance(msg, Image.Image):
        if msg.format:
            format_ = ImageFormat[msg.format.upper()]
        else:
            format_ = ImageFormat.PNG
        return ImageMessage(content=msg, format=format_)

    if isinstance(msg, dict):
        msg_dict: MessageDict = {
            "content": msg["content"],
            "role": msg.get("role", "user"),
            "type": msg.get("type", "text"),
        }

        role = Role(msg_dict["role"])

        if msg_dict["type"] == ContentType.TEXT.value:
            if not isinstance(msg_dict["content"], str):
                raise ValueError("Invalid text message format")
            return TextMessage(content=msg_dict["content"], role=role)
        elif msg_dict["type"] == ContentType.IMAGE.value:
            if not isinstance(msg_dict["content"], (str, Image.Image)):
                raise ValueError("Image content must be either string or PIL.Image")
            return ImageMessage(
                content=msg_dict["content"],
                format=ImageFormat(msg.get("format", "png")),
                detail=ImageDetail(msg.get("detail", "auto")),
            )
        raise ValueError(f"Unsupported content type: {msg_dict['type']}")

    raise ValueError(f"Unsupported message type: {type(msg)}")



# SPDX-FileCopyrightText: 2024 Zsbyqx20 <112002598+Zsbyqx20@users.noreply.github.com>
#
# SPDX-License-Identifier: Apache-2.0

from typing import Any, Callable, Sequence, TypeVar, cast

from openai import AsyncOpenAI, AuthenticationError, OpenAI, PermissionDeniedError
from openai.types.chat import ChatCompletion

from .base import (
    BaseClient,
    BaseMessage,
    MessageInput,
    Role,
    ServiceType,
)
from .message import ImageMessage, TextMessage, convert_to_message
from .utils import async_retry, retry

T = TypeVar("T")


class OpenAIClient(BaseClient):
    def __init__(self, api_key: str | None = None, base_url: str | None = None):
        super().__init__(api_key, base_url, ServiceType.OPENAI)

    def _init_client(self):
        self._client = OpenAI(api_key=self.api_key, base_url=self.base_url)
        self._async_client = AsyncOpenAI(api_key=self.api_key, base_url=self.base_url)

    def _create_openai_content(self, msg: BaseMessage) -> dict[str, Any]:
        if isinstance(msg, TextMessage):
            return {"type": "text", "text": msg.content}
        elif isinstance(msg, ImageMessage):
            return {
                "type": "image_url",
                "image_url": {
                    "url": f"data:image/{msg._format.value};base64,{msg.content}",
                    "detail": msg.detail.value,
                },
            }
        raise ValueError(f"Unsupported message type: {type(msg)}")

    def _merge_user_messages(self, contents: list[dict[str, Any]]) -> dict[str, Any]:
        if not contents:
            return {}

        if all(item["type"] == "text" for item in contents):
            return {
                "role": "user",
                "content": " ".join(item["text"] for item in contents),
            }

        if len(contents) == 1:
            content = contents[0]
            if content["type"] == "text":
                return {"role": "user", "content": content["text"]}
            return {"role": "user", "content": [content]}

        return {"role": "user", "content": contents.copy()}

    def organize_messages(
        self, messages: Sequence[MessageInput]
    ) -> list[dict[str, Any]]:
        processed_messages = [
            msg if isinstance(msg, BaseMessage) else convert_to_message(msg)
            for msg in messages
        ]

        self._validate_message_sequence(processed_messages)

        organized_msgs: list[dict[str, Any]] = []
        current_user_content: list[dict[str, Any]] = []

        def flush_user_content():
            if current_user_content:
                organized_msgs.append(self._merge_user_messages(current_user_content))
                current_user_content.clear()

        for msg in processed_messages:
            if msg.role == Role.USER:
                current_user_content.append(self._create_openai_content(msg))
            else:
                flush_user_content()
                organized_msgs.append({
                    "role": msg.role.value,
                    "content": msg.content
                    if isinstance(msg, TextMessage)
                    else str(msg.content),
                })

        flush_user_content()
        return organized_msgs

    def _process_response(self, response: ChatCompletion) -> tuple[str, Any]:
        resp_obj = response.choices[0].message.content
        resp_usage = response.usage

        if resp_obj is None:
            raise ValueError("No response from OpenAI")
        if resp_usage is None:
            raise ValueError("No usage data from OpenAI")

        return resp_obj, resp_usage

    def _prepare_request(
        self,
        messages: list[MessageInput],
        model: str,
        temperature: float,
        **kwargs: Any,
    ) -> tuple[list[dict[str, Any]], dict[str, Any]]:
        prompt = self.organize_messages(messages)
        params = {
            "model": model,
            "messages": prompt,  # type: ignore
            "temperature": temperature,
            **kwargs,
        }
        return prompt, params

    @retry(max_attempts=3, skip_exceptions=(AuthenticationError, PermissionDeniedError))
    def call(
        self,
        messages: list[MessageInput],
        model: str = "gpt-4o-mini",
        temperature: float = 0.5,
        callback: Callable[[str], T] | None = None,
        **kwargs: Any,
    ):
        _, params = self._prepare_request(messages, model, temperature, **kwargs)

        response = cast(
            ChatCompletion,
            self._client.chat.completions.create(**params),
        )
        resp_obj, _ = self._process_response(response)
        if callback is not None:
            return callback(resp_obj)
        return resp_obj

    @async_retry(
        max_attempts=3, skip_exceptions=(AuthenticationError, PermissionDeniedError)
    )
    async def async_call(
        self,
        messages: list[MessageInput],
        model: str = "gpt-4o-mini",
        temperature: float = 0.5,
        callback: Callable[[str], T] | None = None,
        **kwargs: Any,
    ):
        _, params = self._prepare_request(messages, model, temperature, **kwargs)

        response = await self._async_client.chat.completions.create(**params)
        resp_obj, _ = self._process_response(response)
        if callback is not None:
            return callback(resp_obj)
        return resp_obj



# SPDX-FileCopyrightText: 2024 Zsbyqx20 <112002598+Zsbyqx20@users.noreply.github.com>
#
# SPDX-License-Identifier: Apache-2.0

from typing import Any, Callable, Sequence, TypeVar, cast

from anthropic import Anthropic, AsyncAnthropic
from anthropic._exceptions import AuthenticationError, PermissionDeniedError
from anthropic.types.content_block import ContentBlock
from anthropic.types.message import Message
from anthropic.types.text_block import TextBlock
from anthropic.types.tool_use_block import ToolUseBlock

from .base import (
    BaseClient,
    BaseMessage,
    MessageInput,
    Role,
    ServiceType,
)
from .message import ImageMessage, TextMessage, convert_to_message
from .utils import async_retry, retry

T = TypeVar("T")


def process_content_block_default(block: ContentBlock) -> str:
    if isinstance(block, TextBlock):
        return block.text
    if isinstance(block, ToolUseBlock):
        raise NotImplementedError("Tool use blocks are not supported")
    raise ValueError(f"Unsupported content block type: {type(block)}")


class AnthropicClient(BaseClient):
    def __init__(self, api_key: str | None = None, base_url: str | None = None):
        super().__init__(api_key, base_url, ServiceType.ANTHROPIC)

    def _init_client(self):
        self._client = Anthropic(api_key=self.api_key, base_url=self.base_url)
        self._async_client = AsyncAnthropic(
            api_key=self.api_key, base_url=self.base_url
        )

    def _create_anthropic_content(self, msg: BaseMessage) -> dict[str, Any]:
        if isinstance(msg, TextMessage):
            return {"type": "text", "text": msg.content}
        elif isinstance(msg, ImageMessage):
            return {
                "type": "image",
                "source": {
                    "type": "base64",
                    "media_type": f"image/{msg._format.value}",
                    "data": msg.content,
                },
            }
        raise ValueError(f"Unsupported message type: {type(msg)}")

    def _merge_user_messages(self, contents: list[dict[str, Any]]) -> dict[str, Any]:
        if not contents:
            return {}

        if all(item["type"] == "text" for item in contents):
            return {
                "role": "user",
                "content": " ".join(item["text"] for item in contents),
            }

        if len(contents) == 1:
            content = contents[0]
            if content["type"] == "text":
                return {"role": "user", "content": content["text"]}
            return {"role": "user", "content": [content]}

        return {"role": "user", "content": contents.copy()}

    def organize_messages(
        self, messages: Sequence[MessageInput]
    ) -> list[dict[str, Any]]:
        processed_messages = [
            msg if isinstance(msg, BaseMessage) else convert_to_message(msg)
            for msg in messages
        ]

        self._validate_message_sequence(processed_messages)

        organized_msgs: list[dict[str, Any]] = []
        current_user_content: list[dict[str, Any]] = []

        def flush_user_content():
            if current_user_content:
                organized_msgs.append(self._merge_user_messages(current_user_content))
                current_user_content.clear()

        for msg in processed_messages:
            if msg.role == Role.USER:
                current_user_content.append(self._create_anthropic_content(msg))
            else:
                flush_user_content()
                organized_msgs.append({
                    "role": msg.role.value,
                    "content": msg.content
                    if isinstance(msg, TextMessage)
                    else str(msg.content),
                })

        flush_user_content()
        return organized_msgs

    def _process_response(self, response: Message) -> tuple[ContentBlock, Any]:
        content = response.content
        if not content:
            raise ValueError("No response from Anthropic")

        return content[0], response.usage

    def _prepare_request(
        self,
        messages: list[MessageInput],
        model: str = "claude-3-sonnet-20240229",
        temperature: float = 0.5,
        **kwargs: Any,
    ) -> tuple[list[dict[str, Any]], dict[str, Any]]:
        prompt = self.organize_messages(messages)
        if len(prompt) > 1 and prompt[0]["role"] == "system":
            system_msg = prompt[0]
            prompt = prompt[1:]
        else:
            system_msg = None
        params = {
            "model": model,
            "messages": prompt,
            "temperature": temperature,
            "max_tokens": 4096,
            **kwargs,
        }
        if system_msg:
            params["system"] = system_msg["content"]
        return prompt, params

    @retry(max_attempts=3, skip_exceptions=(AuthenticationError, PermissionDeniedError))
    def call(
        self,
        messages: list[MessageInput],
        model: str = "claude-3-5-haiku-20241022",
        temperature: float = 0.5,
        callback: Callable[[ContentBlock], T] | None = None,
        **kwargs: Any,
    ):
        _, params = self._prepare_request(messages, model, temperature, **kwargs)

        response = cast(Message, self._client.messages.create(**params))
        resp_obj, _ = self._process_response(response)
        if callback is not None:
            return callback(resp_obj)
        return process_content_block_default(resp_obj)

    @async_retry(
        max_attempts=3, skip_exceptions=(AuthenticationError, PermissionDeniedError)
    )
    async def async_call(
        self,
        messages: list[MessageInput],
        model: str = "claude-3-5-haiku-20241022",
        temperature: float = 0.5,
        callback: Callable[[ContentBlock], T] | None = None,
        **kwargs: Any,
    ):
        _, params = self._prepare_request(messages, model, temperature, **kwargs)

        response = cast(Message, await self._async_client.messages.create(**params))
        resp_obj, _ = self._process_response(response)
        if callback is not None:
            return callback(resp_obj)
        return process_content_block_default(resp_obj)
